/*
   Dico

   Luc Maranget, projet Moscova, Inria Rocquencourt
 
   Copyright 2005 Institut National de Recherche en Informatique et    
   en Automatique.  All rights reserved.  This file is distributed     
   under the terms of the Q Public License version 1.0.                
*/

import java.io.* ;

class Dico {
  private int nwords ;
  private int minLen ;
  private int verbose ;
  private int mode ;
  
  private Tst me ;                 // tst is the dictionary
  private Code c ; 
  private Mset m ;                 // multi set used everywhere in scans

  private Words words = null ; // prefix under construction
  private StringBuffer word = new StringBuffer () ; // word idem

  /*
    Lexical enumeration : enumerate words >= pword whose letters
    are taken in m.

    - May call itself to add one word to anagram, then
    pword is the last word found so far. This trick avoids
    duplicates.

    - zyva_all is the whole tst of words to scan, it is static,
    in order to avoid an extra argument on zyva.
                 
  */
  private Tst zyva_all ;  

  
  private void zyva (Tst t, String pword, int depth) {
    if (t == null) return ;
    char pchar = '\0' ;
    if (depth < pword.length()) 
      pchar = pword.charAt(depth) ;

    if (pchar < t.val)
      zyva(t.g, pword, depth) ;

    if (m.here(t.val) && pchar <= t.val) {
      m.decr(t.val) ;
      word.append(t.val) ;
      boolean wordHere = t.count > 0 ;
      boolean prefix = pchar == t.val && depth < pword.length()-1 ;
      if (wordHere && !prefix) {
        String found = words.push(word) ;
        if (found != null) {
          if (verbose > 1) {
            System.err.println("Starting with: " + words) ;
          }
          nwords-- ;
          if (m.isEmpty())
            System.out.println(words) ;
          else if (nwords > 0 && (m.length() >= minLen || words.below() > 0)) {
            zyva(zyva_all, found, 0) ;
          }
          nwords++ ;
          words.pop(word) ;
        }        
      }
      zyva(t.c, pchar == t.val ? pword : "", depth+1) ;
      word.setLength(word.length()-1) ;
      m.incr(t.val) ;
    }
    zyva(t.d, pword, depth) ;
  }



  /*
    Enumeration with one letter c mandatory.
    - t_all is the whole current tst to scan.
    - t is the current subtree of t_all.
    - when a word w is found, register it and then start enumeration
    of t_all filtered by the remaining letters.
    - after, remove w from t_all. (to avoid duplicates).
  */

  private static Tst t_all ;

  private void zyvaWithChar(Tst t, char c) {
    if (t == null) return ;

    zyvaWithChar(t.g, c) ;
    // assert(m.here(t.val)) ; holds because of filtering below
    word.append(t.val) ;
    m.decr(t.val) ;
    if (t.count > 0 && (c == '\0' || c == t.val)) {
      String found = words.push(word) ; // includes length check of word
      if (found != null) {
        if (verbose > 1) {
          System.err.println("Starting with: " + words) ;
        }
        nwords-- ;
        if (m.isEmpty()) {
          System.out.println(words) ;
        } else if (nwords > 0 && (m.length() >= minLen || words.below() > 0)) {
          // if (verbose > 1) System.err.println("Filtering by: " + m) ;
          Tst temp_all = Tst.filter(t_all, m) ;
          //if (verbose > 1) Tst.dump(temp_all, this.c ) ;
          zyvaChoose(temp_all) ;
    // Avoid duplicates, may be omitted when the mandatory letter isn't in m
          t_all = Tst.remove(t_all, found, 0) ;
        }
        nwords++ ;
        words.pop(word) ;
      }
    }
    zyvaWithChar(t.c, c == t.val ? '\0' : c) ;
    m.incr(t.val) ;
    word.setLength(word.length()-1) ;
    zyvaWithChar(t.d, c) ;
  }

  private void zyvaSimple(Tst t) {
    if (t == null) return ;
    //    System.err.println("Simple: " + c.decode(word) + ", " + m + ", " + nwords) ;
    zyvaSimple(t.g) ;
    if (m.here(t.val)) {
      word.append(t.val) ;
      m.decr(t.val) ;
      if (t.count > 0) {
        String found = words.push(word) ;
        if (found != null) {
          if (verbose > 1) {
            System.err.println("Starting with: " + words) ;
          }
          nwords-- ;
          if (m.isEmpty()) {  System.out.println(words) ;}
          else if (nwords > 0 && (m.length() >= minLen || words.below() > 0)) {
            Tst save_all = t_all ;
            zyvaSimple(t_all) ;
            t_all = Tst.remove(save_all, found, 0) ;
          }
          nwords++ ;
          words.pop(word) ;
        }
      }
      zyvaSimple(t.c) ;
      m.incr(t.val) ;
      word.setLength(word.length()-1) ;
    }
    zyvaSimple(t.d) ;
  }
  /*
    Choose either lexical or letter enforced enumeration of tst temp_all.
    In the latter case, the mandatory letter is the least frequent
    in dictionnary.
  */
  private void zyvaChoose(Tst temp_all) {
    if (mode == OOptions.LOWFREQ ||
        (mode == OOptions.NORMAL && m.length() > OOptions.MAXLETTERS)) {
      Tst save_all = t_all ; // avoid an extra t_all argument
      t_all = temp_all ;
      int minChar = Tst.freqs(t_all, c) ;      
      if (minChar != -1) {
        if (verbose > 0) {
          System.err.println("" + words + " -> " + c.decode((char)minChar)) ;
        }
        zyvaWithChar(t_all, (char)minChar) ;
      }
      t_all = save_all ;
    } else {
      zyva_all = temp_all ;
      zyva(zyva_all, "", 0) ;
    }
  }

  void ana() {
    if (verbose > 2) {
      System.err.println("Dictionnaire rduit") ;
      Tst.dump(me, c) ;
    } else if (verbose > 0) {
      System.err.print("words: " + Tst.nwords(me)) ;
      System.err.println(", nodes: " + Tst.nodes(me)) ;
      String pmode = "" ;
      if (mode == OOptions.NORMAL)
        pmode = "normal" ;
      else if (mode == OOptions.LEXICAL)
        pmode = "lexical" ;
      else if (mode == OOptions.LOWFREQ)
        pmode = "lowfreq" ;
      else if (mode == OOptions.SIMPLE)
        pmode = "simple" ;
      else {
        System.err.println("Unkown mode! " + mode) ;
        System.exit(2) ;
      }
      System.err.println("mode: " + pmode) ;
    }
    if (mode == OOptions.SIMPLE) {
      t_all = me ;
      zyvaSimple(me) ;
    } else
      zyvaChoose(me) ;
  }


  Dico(OOptions o, BufferedReader in, Code c, Mset m) {
    this.c = c ;
    this.m = m ;
    this.nwords = o.maxWords ;
    this.verbose = o.verbose ;
    this.mode = o.mode ;

    String line ;
    words = new Words (o.minLen, o.below, c) ;
    try {
      while ((line = in.readLine()) != null) {
        int len = line.length() ;
        //        line = line.toLowerCase() ;
        line = c.code(line) ;
        if ((o.below > 0 || len >= o.minLen) &&
            line != null && m.here(line)) { // insert useful words only
          me = Tst.insert(me, line, 0, len) ;
        }
      }
    } catch (IOException e) {
      System.err.println(e.getMessage()) ;
      System.exit(2) ;
    }
    //    me = Tst.balance(me) ; // does not harm anyway, may not serve either
  }

  
  public static void main(String [] argv) {
    OOptions options = new OOptions("Dico", argv) ;
    options.parse() ;

    Code c = new Code (options.letters) ;
    Mset m = new Mset (options.letters, c) ;
    if (options.verbose > 0) {
      System.err.println("dico: " + options.dico + "") ;
      System.err.println("stock: " + options.letters + "") ;
    }
    try {
      Dico d =
        new Dico(options,
                 new BufferedReader (new FileReader (options.dico)), c, m) ;
      d.ana() ;
    } catch (IOException e) {
      System.err.println(e.getMessage()) ;
      System.exit(2) ;
    }
  }
}
